import logging
import sys
from importlib.util import find_spec


logger = logging.getLogger(__name__)


class WorkerTaskPluginLoaderMixin:
    """Worker Loader Mixin‰

    Separate some logic out into this class

    """

    @property
    def taskImports(self):
        includes = []
        for pluginWorkerMain in list(self._loadedPlugins.values()):
            from peek_plugin_base.server.PluginServerWorkerEntryHookABC import (
                PluginServerWorkerEntryHookABC,
            )

            if isinstance(pluginWorkerMain, PluginServerWorkerEntryHookABC):
                desc = str(pluginWorkerMain).split(" ")[0].strip("<")
                try:
                    imports = pluginWorkerMain.workerTaskImports
                    if imports:
                        logger.debug("Loading tasks for %s", desc)
                        includes.extend(imports)
                    else:
                        logger.debug("No tasks defined for %s", desc)
                except Exception as e:
                    logger.exception(
                        "Failed to load tasks for %s: %s", desc, str(e)
                    )

        return includes


def importWorkerTasks():
    from peek_platform import PeekPlatformConfig

    loader = PeekPlatformConfig.pluginLoader

    assert isinstance(loader, WorkerTaskPluginLoaderMixin), (
        "%s is not a " "WorkerTaskPluginLoaderMixin"
    ) % loader

    for task in loader.taskImports:
        modName, functionName = task.rsplit(".", 1)

        if modName in sys.modules:
            module = sys.modules[modName]
        else:
            modSpec = find_spec(modName)
            if not modSpec or not modSpec.loader:
                raise Exception(
                    f"Failed to find module {modName}, is the python package installed?"
                )

            # Load the module.
            # This will call the decorator and register the task
            modSpec.loader.load_module()



import json
import logging
import socket
from datetime import datetime
from pathlib import Path

import psutil
import pytz
from twisted.internet.task import LoopingCall
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import vortexLogFailure
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_plugin_base.LoopingCallUtil import peekCatchErrbackWithLogger

logger = logging.getLogger(__name__)

# You can replace this prefix with your actual prefix
_osMetricsTuplePrefix = "system.metrics."


@addTupleType
class _OsMetricsTuple(Tuple):
    __tupleType__ = _osMetricsTuplePrefix + "_OsMetricsTuple"

    lastUpdateDate: datetime = TupleField()
    hostFqdn: str = TupleField()
    cpu15MinuteLoadPercent: int = TupleField()
    maxLocalDiskUsagePercent: int = TupleField()
    memoryBufferCacheMb: int = TupleField()
    memoryFreeMb: int = TupleField()
    swapUsageMb: int = TupleField()
    swapUsedPercent: int = TupleField()
    memoryUsedPercent: int = TupleField()
    systemCpuCount: int = TupleField()
    systemMemorySizeMb: int = TupleField()
    systemSwapSizeMb: int = TupleField()
    enabledPlugins: list[str] = TupleField([])
    serviceStartDate: datetime = TupleField()
    serviceUptimeHours: float = TupleField()


class OsMetrics:
    LOOPING_CALL_PERIOD = 30.0
    """Collects and writes system metrics to a JSON file."""

    def __init__(self, writeDir: Path, enabledPlugins: list[str]):
        self._writeDir = writeDir
        self._enabledPlugins = enabledPlugins
        self._serviceStartDate = datetime.now(pytz.utc)
        self._loopingCall = LoopingCall(
            peekCatchErrbackWithLogger(logger)(self.writeMetrics)
        )

    def start(self):
        d = self._loopingCall.start(self.LOOPING_CALL_PERIOD, now=True)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def stop(self):
        if self._loopingCall is not None:
            self._loopingCall.stop()
            self._loopingCall = None

    @deferToThreadWrapWithLogger(logger)
    def writeMetrics(self) -> None:
        """Collect and write system metrics to a JSON file.

        Raises:
            OSError: If there are file permission issues
            psutil.Error: If there are errors collecting system metrics
        """
        try:
            metrics = self._collectMetrics()
            jsonFilePath = self._writeDir / "system_metrics.json"
            hostnameTxtPath = self._writeDir / "hostname.txt"

            # Convert datetime to string for JSON serialization
            metricsDict = metrics.tupleToRestfulJsonDict()

            with jsonFilePath.open("w") as f:
                json.dump(metricsDict, f, indent=4)

            with hostnameTxtPath.open("w") as f:
                f.write(metrics.hostFqdn)

        except Exception as e:
            print(f"Error writing system metrics: {e}")
            raise

    def _getCpuLoadPercent(self) -> int:
        """Get 15-minute CPU load average as a percentage normalized across all CPUs."""
        cpuCount = psutil.cpu_count()
        loadAvg = psutil.getloadavg()[2]  # 15-minute load average
        loadPercent = (loadAvg / cpuCount) * 100
        return int(min(loadPercent, 100))

    def _getMaxDiskUsagePercent(self) -> int:
        """Get maximum disk usage percentage across all mounted filesystems."""
        maxPercent = 0.0
        for partition in psutil.disk_partitions(all=False):
            try:
                usage = psutil.disk_usage(partition.mountpoint)
                maxPercent = int(max(maxPercent, usage.percent))
            except PermissionError:
                continue
        return maxPercent

    def _getMemoryMetrics(self, metrics: _OsMetricsTuple) -> None:
        """Get memory-related metrics including swap usage."""
        virtualMemory = psutil.virtual_memory()
        swap = psutil.swap_memory()

        # Get buffer/cache memory - handle different system reporting methods
        buffersAndCached = getattr(virtualMemory, "buffers", 0) + getattr(
            virtualMemory, "cached", 0
        )
        if buffersAndCached == 0:  # If neither attribute exists
            # On some systems, cached might be calculated from total - used - free - buffers
            buffersAndCached = (
                virtualMemory.total - virtualMemory.used - virtualMemory.free
            )

        metrics.memoryBufferCacheMb = int(buffersAndCached / 1024 / 1024)
        metrics.memoryFreeMb = int(virtualMemory.available / 1024 / 1024)
        metrics.swapUsageMb = int(swap.used / 1024 / 1024)
        metrics.swapUsedPercent = int(swap.percent)
        metrics.memoryUsedPercent = int(virtualMemory.percent)

        # Add the new system memory and swap size metrics
        metrics.systemMemorySizeMb = int(virtualMemory.total / 1024 / 1024)
        metrics.systemSwapSizeMb = int(swap.total / 1024 / 1024)

    def _collectMetrics(self) -> _OsMetricsTuple:
        """Collect all system metrics and return as an OsMetricsTuple."""
        hours = round(
            int(
                (
                    datetime.now(pytz.utc) - self._serviceStartDate
                ).total_seconds()
                / 60
            )
            / 60,
            4,
        )

        metrics = _OsMetricsTuple(
            hostFqdn=socket.gethostname(),
            cpu15MinuteLoadPercent=self._getCpuLoadPercent(),
            maxLocalDiskUsagePercent=self._getMaxDiskUsagePercent(),
            lastUpdateDate=datetime.now(pytz.utc),
            systemCpuCount=psutil.cpu_count(),
            enabledPlugins=self._enabledPlugins,
            serviceStartDate=self._serviceStartDate,
            serviceUptimeHours=hours,
        )
        self._getMemoryMetrics(metrics)
        return metrics



import json
import logging
import os
import sys

from peek_platform.subproc_plugin_init.plugin_subproc_parent_main import (
    PluginSubprocParentMain,
)
from peek_plugin_base.PeekVortexUtil import peekBackendNames
from sqlalchemy.util import b64encode
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.Tuple import Tuple

from peek_platform.subproc_plugin_init.plugin_subproc import (
    plugin_subproc_child_main,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    LOGGING_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_parent_protocol import (
    PluginSubprocParentProtocol,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_platform_config_tuple import (
    PluginSubprocPlatformConfigTuple,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_vortex_payload_envelope_tuple import (
    PluginSubprocVortexPayloadEnvelopeTuple,
)
from peek_plugin_base.PeekPlatformCommonHookABC import PeekPlatformCommonHookABC
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_base.PluginCommonEntryHookABC import PluginCommonEntryHookABC


logger = logging.getLogger(__name__)


class _NoKeycheckPayloadEndpoint(PayloadEndpoint):
    def _keyCheck(self, filt):
        pass


class PluginSubprocParentMainDelegate(PluginCommonEntryHookABC):
    def __init__(
        self,
        pluginName: str,
        pluginRootDir: str,
        platform: PeekPlatformCommonHookABC,
        subprocessGroupMain: PluginSubprocParentMain,
    ):
        super().__init__(pluginName, pluginRootDir)

        assert "-" not in pluginName, "Plugin name must not have hyphens"
        self._pluginName = pluginName
        self._platform = platform
        self._subprocessGroupMain = subprocessGroupMain

        from peek_platform import PeekPlatformConfig

        self._serviceName = PeekPlatformConfig.componentName

        self._pluginEndpoint = None

    @inlineCallbacks
    def load(self) -> None:
        yield self._subprocessGroupMain.sendPluginLoad(self._pluginName)
        logger.debug("Loaded Standalone Plugin %s", self._pluginName)

    @inlineCallbacks
    def start(self) -> None:
        from peek_platform import PeekPlatformConfig

        yield self._subprocessGroupMain.sendPluginStart(self._pluginName)

        self._pluginEndpoint = _NoKeycheckPayloadEndpoint(
            dict(plugin=self._pluginName),
            self._subprocessGroupMain.sendPayloadEnvelopeToChild,
            ignoreFromVortex=(peekServerName, PeekPlatformConfig.componentName),
        )

        logger.debug("Started Standalone Plugin %s", self._pluginName)

    @inlineCallbacks
    def stop(self) -> None:
        yield None

        if self._pluginEndpoint:
            yield self._pluginEndpoint.shutdown()
            self._pluginEndpoint = None

        # yield self._subprocessGroupMain.sendPluginStop(self._pluginName)
        # logger.debug("Stopped Standalone Plugin %s", self._pluginName)
        logger.debug(
            "Standalone Plugin %s doesn't support stopping", self._pluginName
        )

    @inlineCallbacks
    def unload(self) -> None:
        yield None
        # yield self._subprocessGroupMain.sendPluginUnload(self._pluginName)
        # logger.debug("Unloaded Standalone Plugin %s", self._pluginName)
        logger.debug(
            "Standalone Plugin %s doesn't support unloading", self._pluginName
        )



import logging
import shlex
import subprocess
from typing import Optional
from typing import Tuple

from twisted.internet import reactor
from twisted.internet import threads
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from twisted.internet.protocol import ProcessProtocol
from vortex.DeferUtil import vortexLogFailure

logger = logging.getLogger(__name__)


class SpawnOsCommandException(subprocess.CalledProcessError):
    """Spawn OS Command Exception

    This exception is raised when an OS command fails or returns a non-zero exit code

    """

    def __init__(
        self,
        returncode: int,
        cmd: str,
        stdout: Optional[str] = None,
        stderr: Optional[str] = None,
        message: Optional[str] = None,
    ):
        """Constructor

        :param returncode: The code returned by the process.
        :param cmd: A string representing the command and args executed by bash.
        :param stdout: The output from the process.
        :param stderr: The error from the process.
        :param message: An additional message that can be used to emulate the message
        method of standard exceptions
        """

        subprocess.CalledProcessError.__init__(
            self, returncode, cmd, stdout, stderr
        )
        self.message = message

    def __str__(self):
        return "%s\nCommand '%s' returned non-zero exit status %d" % (
            self.message,
            self.cmd,
            self.returncode,
        )


def spawnSubprocessBlocking(
    cmdAndArgs: str,
    logger_: logging.Logger,
    cwd: Optional[str] = None,
    env: Optional[dict] = None,
    exceptionOnStderr: bool = True,
    exceptionOnNonZeroExitCode: bool = True,
    stderrLogLevel: int = logging.ERROR,
) -> Tuple[bytes, bytes, int]:
    return threads.blockingCallFromThread(
        reactor,
        spawnSubprocess,
        cmdAndArgs,
        logger_,
        cwd=cwd,
        env=env,
        exceptionOnStderr=exceptionOnStderr,
        exceptionOnNonZeroExitCode=exceptionOnNonZeroExitCode,
        stderrLogLevel=stderrLogLevel,
    )


@inlineCallbacks
def spawnSubprocess(
    cmdAndArgs: str,
    logger_: logging.Logger,
    cwd: Optional[str] = None,
    env: Optional[dict] = None,
    exceptionOnStderr: bool = True,
    exceptionOnNonZeroExitCode: bool = True,
    stderrLogLevel: int = logging.ERROR,
) -> Deferred:
    """Spawn Subprocess

    This method calls an OS command using the subprocess package and bash as the
    interpreter.

    :param cmdAndArgs: A string containing the command
        and arguments to pass to bash.
    :param logger_: The parser for the command output.
    :param env: os.environ to use for the command
    :param exceptionOnStderr: Should the deferred errback if there is stdout?

    :returns: None, This either succeeds or raises an exception.

    :raises: c{CalledProcessError}

    """
    assert isinstance(cmdAndArgs, str), "cmdAndArgs is not str"

    shellCmdAndArgs = """ bash -l -c "cd '%s' && %s" """ % (cwd, cmdAndArgs)
    args = shlex.split(shellCmdAndArgs)
    cmd = args[0]
    pp = ProcessOutLogger(
        shellCmdAndArgs,
        logger_,
        exceptionOnStderr,
        exceptionOnNonZeroExitCode,
        stderrLogLevel=stderrLogLevel,
    )
    # noinspection PyUnresolvedReferences
    logger_.debug("Running command '%s'", shellCmdAndArgs)
    reactor.spawnProcess(pp, cmd, args, env=env, path=cwd)

    pp.deferred.addErrback(vortexLogFailure, logger_, consumeError=False)
    return (yield pp.deferred)


class ProcessOutLogger(ProcessProtocol):
    def __init__(
        self,
        command: str,
        logger_: logging.Logger,
        exceptionOnStderr: bool,
        exceptionOnNonZeroExitCode: bool,
        stderrLogLevel: int,
    ):
        self._command = command
        self._logger = logger_
        self._exceptionOnStderr = exceptionOnStderr
        self._exceptionOnNonZeroExitCode = exceptionOnNonZeroExitCode
        self._stderrLogLevel = stderrLogLevel
        self.deferred = Deferred()
        self.stdout = b""
        self.stderr = b""

    def connectionMade(self):
        self._logger.debug("connectionMade, command=%s" % (self._command,))
        self.transport.closeStdin()

    def outReceived(self, data):
        self.stdout += data
        for line in data.splitlines():
            if not line.decode().strip():
                continue
            self._logger.debug(line.decode().strip())

    def errReceived(self, data):
        self.stderr += data
        for line in data.splitlines():
            if not line.decode().strip():
                continue
            self._logger.log(self._stderrLogLevel, line.decode().strip())

    def processEnded(self, reason):
        if self._exceptionOnStderr and self.stderr:
            self.deferred.errback(
                Exception(
                    "Command generated, exit code %s,"
                    " stderr: %s"
                    % (reason.value.exitCode, self.stderr.decode())
                )
            )

        elif self._exceptionOnNonZeroExitCode and reason.value.exitCode:
            self.deferred.errback(
                Exception(
                    "Command exited with non-zero exit code %s,"
                    " stderr: %s"
                    % (reason.value.exitCode, self.stderr.decode())
                )
            )

        else:
            self.deferred.callback(
                (self.stdout, self.stderr, reason.value.exitCode)
            )

        self._logger.debug(
            "processEnded, status %d, len(stdout) %s, len(stderr) %s, cmd: %s"
            % (
                reason.value.exitCode,
                len(self.stdout),
                len(self.stderr),
                self._command,
            )
        )



import json
import logging
import sys
from base64 import b64decode

from setproctitle import setproctitle
from twisted.internet import reactor
from twisted.internet._posixstdio import StandardIO
from vortex.VortexFactory import VortexFactory

from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_child_state_protocol import (
    PluginSubprocChildStateProtocol,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_child_vortex_protocol import (
    PluginSubprocChildVortexProtocol,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_child_vortex_uuid_protocol import (
    PluginSubprocChildVortexUuidProtocol,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_platform_config_tuple import (
    PluginSubprocPlatformConfigTuple,
)


"""
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_platform_config_tuple import PluginSubprocPlatformConfigTuple
from base64 import b64encode
import json

x=PluginSubprocPlatformConfigTuple(serviceName="peek-office-service", pluginName="peek_core_search")
b64encode(json.dumps(x.toJsonDict()).encode())

b'eyJfY3QiOiJydCIsIl9jIjoicGVla19wbGF0Zm9ybS5QbHVnaW5TdWJwcm9jUGxhdGZvcm1Db25maWdUdXBsZSIsInBsdWdpbk5hbWUiOiJwZWVrX2NvcmVfc2VhcmNoIiwic2VydmljZU5hbWUiOiJwZWVrLW9mZmljZS1zZXJ2aWNlIn0='
"""

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
    base64Tuple = sys.argv[1]

    platformConfigTuple = PluginSubprocPlatformConfigTuple().fromJsonDict(
        json.loads(b64decode(base64Tuple))
    )
    logger = logging.getLogger(
        "subproc plugin main %s %s"
        % (platformConfigTuple.serviceName, platformConfigTuple.subprocessGroup)
    )

    setproctitle(
        "%s %s"
        % (platformConfigTuple.serviceName, platformConfigTuple.subprocessGroup)
    )

    from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_child_vortex import (
        PluginSubprocChildVortex,
    )

    vortex = PluginSubprocChildVortex(
        vortexName=platformConfigTuple.serviceName
    )

    # Create the protocol used to send vortex messages
    sendVortexToParentProtocol = PluginSubprocChildVortexProtocol(vortex)
    StandardIO(
        sendVortexToParentProtocol,
        stdin=VORTEX_MSG_TO_CHILD_FD,
        stdout=VORTEX_MSG_FROM_CHILD_FD,
    )

    try:
        # If we run this directly from the commandline, FD3 and FD4 don't exist
        # Just log the problem
        # This is the vortex UUID update protocol
        StandardIO(
            PluginSubprocChildVortexUuidProtocol(vortex),
            stdin=VORTEX_UUID_TO_CHILD_FD,
            stdout=VORTEX_UUID_FROM_CHILD_FD,
        )

        # This is the plugin load, start, stop, unload protocol
        StandardIO(
            PluginSubprocChildStateProtocol(
                platformConfigTuple.serviceName,
                platformConfigTuple.subprocessGroup,
            ),
            stdin=PLUGIN_STATE_TO_CHILD_FD,
            stdout=PLUGIN_STATE_FROM_CHILD_FD,
        )

    except Exception as e:
        logger.error("Failed to connect to FD3 and FD4")
        logger.exception(e)

    vortex.setStdoutProtocol(sendVortexToParentProtocol)
    VortexFactory.addCustomServerVortex(vortex)

    reactor.run()
